#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 14 21:31:31 2023

@author: everest
"""

import requests
from pathlib import Path
import typer

def kgml(species, results):
    """
    Handles the API call for acquiring the specific species input
    """
    KEGGorg = 'http://rest.kegg.jp/list/organism'
    KEGGlist = 'http://rest.kegg.jp/list/pathway/%s'
    KEGGget = 'http://rest.kegg.jp/get/%s/kgml'
    response = requests.get(KEGGorg).text.split('\n')
    org_list = []
    taxonomy = []
    # The -1 avoids the last element of the list, which is a dangling newline
    for i in range(len(response) - 1):
        org_list.append(response[i].split('\t')[1])
        taxonomy.append(response[i].split('\t')[2])
    d = {org_list[i]: taxonomy[i] for i in range(len(taxonomy))}
    if species not in org_list:
        typer.echo(f'Please input a species name in KEGG organism ID format.\nThese are usually {len(min(org_list, key = len))} to {len(max(org_list, key = len))} letter codes.\n--Example: "Homo sapiens" is "hsa"')
    else:
        typer.echo(f'Now acquiring all KGML files for {d[species]}...')
        response = requests.get(KEGGlist % species).text.split('\n')
        pathways = [r.split('\t')[0] for r in response if r]
        for path in pathways:
            typer.echo(f'Now acquiring pathway {path}...')
            config = Path(results / '{}.xml'.format(path))
            paths = requests.get(KEGGget % path).text
            if config.is_file():
                pass
            else:
                with open(results / '{}.xml'.format(path), 'w') as outfile:
                    outfile.write(paths)
            


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan 31 20:31:40 2023

@author: everest_castaneda1
"""

import pytest
import requests
from pathlib import Path
import xml.etree.ElementTree as ET

def test_get_kgml():
    KEGGget = 'http://rest.kegg.jp/get/hsa00232/kgml'
    response = requests.get(KEGGget).text
    root = ET.fromstring(response)
    
    tree_data = ET.parse('data/hsa00232.xml')
    root_data = tree_data.getroot()
    
    assert root_data.get('name') == root.get('name')
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 29 20:49:11 2023

@author: everest_castaneda1
"""

import sys
import typer
import click
from pathlib import Path

from .knext.convert import convert_file
from .knext.convert import convert_folder
from .knext.call import kgml
from .knext.genes import genes_file
from .knext.genes import genes_folder
from .knext.mixed import mixed_file
from .knext.mixed import mixed_folder

@click.group()
def cli():
    pass

@cli.command()
@click.argument('species')
@click.option('-r', '--results', help = 'Directory in which to save output', required = False)
def get_kgml(species, results):
    """
    Acquires all KGML files for a given species. Use a KEGG species
    identification, usually a 3 to 4 letter organism code, as input. Handles
    directories first, hence make sure you have a proper species identifier or
    it may return empty, junk folders. The results flag is to save all results 
    to a directory. For more information about KEGG organism codes, visit: 
    https://www.genome.jp/kegg/catalog/org_list.html
    """
    if results is not None:
        results = Path(results)
        if results.exists() == False:
            typer.echo(f'Directory {results} does not exist or is invalid. Please input a valid directory...')
            sys.exit()
        else:
            kgml(species, results)
    else:
        wd = Path.cwd()
        results = wd / 'kgml_{}'.format(species)
        typer.echo(f'No output directory provided. All files will be saved to:\n{results}')
        results.mkdir(exist_ok = True)
        kgml(species, results)
        
@cli.command()
@click.argument('input_data')
@click.option('-r', '--results', required = False)
@click.option('-u', '--unique', default = False, is_flag = True)
@click.option('-g', '--graphics', default = False, is_flag = True)
def genes(input_data: str, results: str, unique: bool, graphics: bool):
    """
    Converts a folder of KGML files or a single KGML file into a weighted
    edgelist of genes that can be used in graph analysis. If -u/--unique flag 
    is used genes are returned with terminal modifiers to enhance network 
    visualization or analysis. If the -g/--graphics flag is used, x-y 
    coordinates of pathways are returned, which may be used as positions in 
    NetworkX\'s graph drawing commands.
    """
    if Path.exists(Path(input_data)) == False:
        typer.echo('Please input a directory of KGML files or an individual KGML file...')
        sys.exit()
    else:
        if results:
            if Path.exists(Path(results)) == False:
                typer.echo('Directory not found.\nPlease input a directory in which to save the results...')
            else:
                wd = Path(results)
                if graphics and unique:
                    if Path(input_data).is_file():
                        genes_file(input_data, wd, unique = True, graphics = True)
                    else:
                        genes_folder(input_data, wd, unique = True, graphics = True)
                elif not graphics and unique:
                    if Path(input_data).is_file():
                        genes_file(input_data, wd, unique = True)
                    else:
                        genes_folder(input_data, wd, unique = True)
                elif graphics and not unique:
                    if Path(input_data).is_file():
                        genes_file(input_data, wd, graphics = True)
                    else:
                        genes_folder(input_data, wd, graphics = True)
                else:
                    if Path(input_data).is_file():
                        genes_file(input_data, wd)
                    else:
                        genes_folder(input_data, wd)
                    
        else:
            wd = Path.cwd()
            typer.echo(f'\nNo output directory given. All resulting files or folders will be saved to current directory:\n{wd}\n')
            if graphics and unique:
                if Path(input_data).is_file():
                    genes_file(input_data, wd, unique = True, graphics = True)
                else:
                    genes_folder(input_data, wd, unique = True, graphics = True)
            elif not graphics and unique:
                if Path(input_data).is_file():
                    genes_file(input_data, wd, unique = True)
                else:
                    genes_folder(input_data, wd, unique = True)
            elif graphics and not unique:
                if Path(input_data).is_file():
                    genes_file(input_data, wd, graphics = True)
                else:
                    genes_folder(input_data, wd, graphics = True)
            else:
                if Path(input_data).is_file():
                    genes_file(input_data, wd)
                else:
                    genes_folder(input_data, wd)

@cli.command()
@click.argument('input_data')
@click.option('-r', '--results', required = False)
@click.option('-u', '--unique', default = False, is_flag = True)
@click.option('-g', '--graphics', default = False, is_flag = True)
def mixed(input_data: str, results: str, unique: bool = False, graphics: bool = False):
    """
    Converts a folder of KGML files or a single KGML file into a weighted
    edgelist of mixed genes, compounds, and pathways that can be used in graph 
    analysis. If -u/--unique flag is used nodes are returned with terminal 
    modifiers to enhance network visualization or analysis. If the 
    -g/--graphics flag is used, x-y coordinates of pathways are returned, 
    which may be used as positions in NetworkX\'s graph drawing commands.
    """
    if Path.exists(Path(input_data)) == False:
        typer.echo('Please input a directory of KGML files or an individual KGML file...')
        sys.exit()
    else:
        if results:
            if Path.exists(Path(results)) == False:
                typer.echo('Directory not found.\nPlease input a directory in which to save the results...')
            else:
                wd = Path(results)
                if graphics and unique:
                    if Path(input_data).is_file():
                        mixed_file(input_data, wd, unique = True, graphics = True)
                    else:
                        mixed_folder(input_data, wd, unique = True, graphics = True)
                elif not graphics and unique:
                    if Path(input_data).is_file():
                        mixed_file(input_data, wd, unique = True)
                    else:
                        mixed_folder(input_data, wd, unique = True)
                elif graphics and not unique:
                    if Path(input_data).is_file():
                        mixed_file(input_data, wd, graphics = True)
                    else:
                        mixed_folder(input_data, wd, graphics = True)
                else:
                    if Path(input_data).is_file():
                        mixed_file(input_data, wd)
                    else:
                        mixed_folder(input_data, wd)
        else:
            wd = Path.cwd()
            typer.echo(f'No output directory given. All resulting files or folders will be saved to current directory:\n{wd}\n')
            if graphics and unique:
                if Path(input_data).is_file():
                    mixed_file(input_data, wd, unique = True, graphics = True)
                else:
                    mixed_folder(input_data, wd, unique = True, graphics = True)
            elif not graphics and unique:
                if Path(input_data).is_file():
                    mixed_file(input_data, wd, unique = True)
                else:
                    mixed_folder(input_data, wd, unique = True)
            elif graphics and not unique:
                if Path(input_data).is_file():
                    mixed_file(input_data, wd, graphics = True)
                else:
                    mixed_folder(input_data, wd, graphics = True)
            else:
                if Path(input_data).is_file():
                    mixed_file(input_data, wd)
                else:
                    mixed_folder(input_data, wd)
        
@cli.command()
@click.argument('species')
@click.argument('input_data')
@click.option('-r', '--results', required = False)
@click.option('-u', '--unique', default = False, is_flag = True)
@click.option('-up', '--uniprot', default = False, is_flag = True)
@click.option('-g', '--graphics', required = False)
def convert(input_data, species, graphics, results: bool = False, uniprot: bool = False, unique: bool = False):
    """
    Converts a file or folder of parsed genes or mixed pathways from KEGG IDs to 
    NCBI gene IDs or UniProt IDs. Default is NCBI gene IDs unless the
    -up/--uniprot flag is used. The -g/--graphics flag converts the graphics file
    or folder into the accompanying IDs. Do not mix/match! Use the -u/--unique 
    flag only if the input is unique, presence of terminal modifiers. 
    Use the -g/--graphics flag with folders if the input for the conversion is 
    a folder and vice-versa.
    """
    if Path.exists(Path(input_data)) == False:
        typer.echo('Please input a directory of KGML files or an individual KGML file...')
        sys.exit()
    else:
        if results:
            wd = Path(results)
            if graphics and unique and uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, graphics, unique = True, uniprot = True)
                else:
                    convert_folder(species, input_data, wd, graphics, unique = True, uniprot = True)
            elif graphics and not unique and uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, graphics, uniprot = True)
                else:
                    convert_folder(species, input_data, wd, graphics, uniprot = True)
            elif graphics and unique and not uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, graphics, unique = True)
                else:
                    convert_folder(species, input_data, wd, graphics, unique = True)
            elif graphics and not unique and not uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, graphics)
                else:
                    convert_folder(species, input_data, wd, graphics)
            elif not graphics and unique and uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, unique = True, uniprot = True)
                else:
                    convert_folder(species, input_data, wd, unique = True, uniprot = True)
            elif not graphics and not unique and uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, uniprot = True)
                else:
                    convert_folder(species, input_data, wd, uniprot = True)
            elif not graphics and unique and not uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, unique = True)
                else:
                    convert_folder(species, input_data, wd, unique = True)
            else:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd)
                else:
                    convert_folder(species, input_data, wd)       
        else:
            wd = Path.cwd()
            typer.echo(f'\nNo output directory given. All resulting files or folders will be saved to current directory:\n{wd}\n')
            if graphics and unique and uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, graphics, unique = True, uniprot = True)
                else:
                    convert_folder(species, input_data, wd, graphics, unique = True, uniprot = True)
            elif graphics and not unique and uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, graphics, uniprot = True)
                else:
                    convert_folder(species, input_data, wd, graphics, uniprot = True)
            elif graphics and unique and not uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, graphics, unique = True)
                else:
                    convert_folder(species, input_data, wd, graphics, unique = True)
            elif graphics and not unique and not uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, graphics)
                else:
                    convert_folder(species, input_data, wd, graphics)
            elif not graphics and unique and uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, unique = True, uniprot = True)
                else:
                    convert_folder(species, input_data, wd, unique = True, uniprot = True)
            elif not graphics and not unique and uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, uniprot = True)
                else:
                    convert_folder(species, input_data, wd, uniprot = True)
            elif not graphics and unique and not uniprot:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd, unique = True)
                else:
                    convert_folder(species, input_data, wd, unique = True)
            else:
                if Path(input_data).is_file():
                    convert_file(species, input_data, wd)
                else:
                    convert_folder(species, input_data, wd)



if __name__ == '__main__':
    cli()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Everest Uriel Castaneda
@desc: File for obtaining mixed pathway output
"""

import json
import typer
import pandas as pd
from pathlib import Path
from itertools import combinations
import xml.etree.ElementTree as ET
from collections import defaultdict

def conv_dict(root):
    # This dictionary is the default option
    # Only compounds are unique
    entry_dict = defaultdict(list)
    for entries in root.findall('entry'):
        for key, items in entries.attrib.items():
            entry_dict[key].append(items)  
       
    entry_id=[]
    entry_name=[]
    for key, items in entry_dict.items():
        if key == 'id':
            for i in items:
                entry_id.append(i)
        if key == 'name':
            for i in items:
                entry_name.append(i)

    unique_compound = []
    for i in range(0, len(entry_id)):
        c = [name + '-' + entry_id[i] if name.startswith('cpd:') or name == 'undefined' else name for name in entry_name[i].split()]
        unique_compound.append(' '.join(c))

    conversion_dictionary = dict(zip(entry_id, unique_compound))
    return conversion_dictionary

def conv_dict_unique(root):
    # This dictionary is the unique version
    # Every item is unique to reveal subgraphs
    entry_dict = defaultdict(list)
    for entries in root.findall('entry'):
        for key, items in entries.attrib.items():
            entry_dict[key].append(items)  
       
    entry_id=[]
    entry_name=[]
    for key, items in entry_dict.items():
        if key == 'id':
            for i in items:
                entry_id.append(i)
        if key == 'name':
            for i in items:
                entry_name.append(i)
   
    unique_name = []
    for i in range(0, len(entry_id)):
        e = [name + '-' + entry_id[i] for name in entry_name[i].split()]
        unique_name.append(' '.join(e))

    conversion_dictionary = dict(zip(entry_id, unique_name))
    return conversion_dictionary


def graphics_dict(root):
    entry_dict = defaultdict(list)
    for entry in root.findall('entry'):
        for graphics in entry.find('graphics').items():
            if graphics[0] == 'x' or graphics[0] == 'y':
                entry_dict[entry.attrib['id']].append(int(graphics[1]))
    graphics_dict = {}
    for key, items in entry_dict.items():
        graphics_dict[key] = tuple(items)
    return graphics_dict

def mixed_file(input_data: str, wd: Path, unique: bool = False, graphics: bool = False):
    """
    Converts a folder of KGML files or a single KGML file into a weighted
    edgelist of mixed genes, compounds, and pathways that can be used in graph 
    analysis. If -u/--unique flag is used nodes are returned with terminal 
    modifiers to enhance network visualization or analysis. If the 
    -g/--graphics flag is used, x-y coordinates of pathways are returned, 
    which may be used as positions in NetworkX\'s graph drawing commands.
    """
    file = Path(input_data)
    tree = ET.parse(file)
    root = tree.getroot()
   
    title = root.get('title')
    pathway = root.get('name').replace('path:', '')
    pathway_link = root.get('link')
    if graphics:
        # Graphics
        graphics = graphics_dict(root)
        
        typer.echo(f'Now parsing: {title}...')
           
        e1 = []
        e2 = []
        n = []
        v = []
        for relation in root.findall('relation'):
            e1.append(relation.get('entry1'))
            e2.append(relation.get('entry2'))
            for subtype in relation.findall('subtype'):
                n.append(subtype.get('name'))
                v.append(subtype.get('value'))
           
        d = []
        for relation in root.findall('relation'):
            for subtype in relation:
                d1=relation.attrib
                d2=subtype.attrib
                d3=json.dumps(d1),json.dumps(d2)
                d.append(d3)    

        edgelist=[]
        # Takes json dictionary and removes formatting in order to obtain an edgelist
        # This odd line of code is necessary  because of the various nested and non-nested subtypes
        # This allows for edges to inherit weights from a nested subtype
        for line in d:
            x=line[0].replace("{","").replace("}","").replace('"entry1":',"")\
                .replace('"entry2":',"").replace('"type":',"").replace('"name":',"")\
                .replace('"value":',"").replace('"','').replace(',',"\t").replace(' ', '')
            y=line[1].replace("{","").replace("}","").replace('"name":',"")\
                .replace('"',"").replace('value:',"").replace(",",'\t').replace(' ', '')
            edgelist.append(x+"\t"+y)

        df=pd.DataFrame(edgelist)
       
        if df.empty:
            typer.echo(f'File "{file}" cannot be parsed.\nVisit {pathway_link} for pathway details.\nThere are likely no edges in which to parse...')
        else:
            df=df[0].str.split("\t", expand=True).rename({0: 'entry1',1: 'entry2',
                                                      2: 'types', 3:'name',
                                                      4: 'value'}, axis='columns')

            entry_dict = defaultdict(list)
            for entries in root.findall('entry'):
                for key, items in entries.attrib.items():
                    entry_dict[key].append(items) 
            
            # Graphics
            df['pos1'] = df['entry1'].map(graphics)
            df['pos2'] = df['entry2'].map(graphics)
           
            entry_id=[]
            entry_name=[]
            entry_type=[]
            for key, items in entry_dict.items():
                if key == 'id':
                    for i in items:
                        entry_id.append(i)
                if key == 'name':
                    for i in items:
                        entry_name.append(i)
                if key == 'type':
                    for i in items:
                        entry_type.append(i)
            
            if unique:
                conversion_dictionary = conv_dict_unique(root)
            else:
                conversion_dictionary = conv_dict(root)
           
            df['entry1'] = df['entry1'].map(conversion_dictionary)
            df['entry2'] = df['entry2'].map(conversion_dictionary)
            df['entry1'] = df['entry1'].astype(str).str.split(' ', expand = False)
            df['entry2'] = df['entry2'].astype(str).str.split(' ', expand = False)
           
            # Parse the cliques seperate so they won't inherit neighbor weights
            # Allows for custom weights so user knows which are customly parsed
            clique_edges = []
            for index, rows in df.iterrows():
                if len(rows['entry1']) > 1:
                    cliques = [x for x in combinations(rows['entry1'], 2)]
                    cliques1 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                    clique_edges.append(cliques1)
                if len(rows['entry2']) > 1:
                    cliques = [x for x in combinations(rows['entry2'], 2)]
                    cliques2 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                    clique_edges.append(cliques2)
            clique2df = [e for edge in clique_edges for e in edge]
            cliquedf = pd.DataFrame.from_records(clique2df, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
           
            edges = []
            for index, rows in df.iterrows():
                lists = rows['entry1'] + rows['entry2']
                # Cliques here inherit neighbor weights, but will be overwritten by above
                cliques = [x for x in combinations(lists, 2)]
                network = [tup + (rows['types'], rows['value'], rows['name'], rows['pos1'], rows['pos2']) for tup in cliques]
                edges.append(network)
            # Standard line for obtaining a dataframe of edges
            edges2df = [e for edge in edges for e in edge]
            # Create pandas DF from edges
            df_out = pd.DataFrame.from_records(edges2df, columns = ['entry1', 'entry2', 'type', 'value', 'name', 'pos1', 'pos2'])
            
            # graphics
            pos_dict1 = {}
            pos_dict2 = {}
            for index, rows in df_out.iterrows():
                pos_dict1[rows['entry1']] = rows['pos1']
                pos_dict2[rows['entry2']] = rows['pos2']
            pos = pos_dict1 | pos_dict2
            json_dict = json.dumps(pos)
            with open(wd / '{}_graphics.txt'.format(Path(graphics).name), 'w') as outfile:
                outfile.write(json_dict)
            
            dft = df_out.groupby(['entry1', 'entry2'])['type'].apply(list).reset_index()
            dfv = df_out.groupby(['entry1', 'entry2'])['value'].apply(list).reset_index()
            dfn = df_out.groupby(['entry1', 'entry2'])['name'].apply(list).reset_index()
            dfx = dft
            dfx['value'] = dfv['value'].agg(','.join)
            dfx['name'] = dfn['name'].agg(','.join)
            dfx['type'] = dft['type'].agg(','.join)
            # Ensures independently parsed cliques overwrite the cliques, which inherited neighbor weights
            xdf = pd.concat([dfx, cliquedf]).drop_duplicates(subset = ['entry1', 'entry2'], keep = 'last')
            xdf.to_csv(wd / '{}.tsv'.format(pathway), sep = '\t', index = False)
    else:
        typer.echo(f'Now parsing: {title}...')
           
        e1 = []
        e2 = []
        n = []
        v = []
        for relation in root.findall('relation'):
            e1.append(relation.get('entry1'))
            e2.append(relation.get('entry2'))
            for subtype in relation.findall('subtype'):
                n.append(subtype.get('name'))
                v.append(subtype.get('value'))
           
        d = []
        for relation in root.findall('relation'):
            for subtype in relation:
                d1=relation.attrib
                d2=subtype.attrib
                d3=json.dumps(d1),json.dumps(d2)
                d.append(d3)    

        edgelist=[]
        # Takes json dictionary and removes formatting in order to obtain an edgelist
        # This odd line of code is necessary  because of the various nested and non-nested subtypes
        # This allows for edges to inherit weights from a nested subtype
        for line in d:
            x=line[0].replace("{","").replace("}","").replace('"entry1":',"")\
                .replace('"entry2":',"").replace('"type":',"").replace('"name":',"")\
                .replace('"value":',"").replace('"','').replace(',',"\t").replace(' ', '')
            y=line[1].replace("{","").replace("}","").replace('"name":',"")\
                .replace('"',"").replace('value:',"").replace(",",'\t').replace(' ', '')
            edgelist.append(x+"\t"+y)

        df=pd.DataFrame(edgelist)
       
        if df.empty:
            typer.echo(f'File "{file}" cannot be parsed.\nVisit {pathway_link} for pathway details.\nThere are likely no edges in which to parse...')
        else:
            df=df[0].str.split("\t", expand=True).rename({0: 'entry1',1: 'entry2',
                                                      2: 'types', 3:'name',
                                                      4: 'value'}, axis='columns')

            entry_dict = defaultdict(list)
            for entries in root.findall('entry'):
                for key, items in entries.attrib.items():
                    entry_dict[key].append(items)
           
            entry_id=[]
            entry_name=[]
            entry_type=[]
            for key, items in entry_dict.items():
                if key == 'id':
                    for i in items:
                        entry_id.append(i)
                if key == 'name':
                    for i in items:
                        entry_name.append(i)
                if key == 'type':
                    for i in items:
                        entry_type.append(i)
            
            if unique:
                conversion_dictionary = conv_dict_unique(root)
            else:
                conversion_dictionary = conv_dict(root)
           
            df['entry1'] = df['entry1'].map(conversion_dictionary)
            df['entry2'] = df['entry2'].map(conversion_dictionary)
            df['entry1'] = df['entry1'].astype(str).str.split(' ', expand = False)
            df['entry2'] = df['entry2'].astype(str).str.split(' ', expand = False)
           
            # Parse the cliques seperate so they won't inherit neighbor weights
            # Allows for custom weights so user knows which are customly parsed
            clique_edges = []
            for index, rows in df.iterrows():
                if len(rows['entry1']) > 1:
                    cliques = [x for x in combinations(rows['entry1'], 2)]
                    cliques1 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                    clique_edges.append(cliques1)
                if len(rows['entry2']) > 1:
                    cliques = [x for x in combinations(rows['entry2'], 2)]
                    cliques2 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                    clique_edges.append(cliques2)
            clique2df = [e for edge in clique_edges for e in edge]
            cliquedf = pd.DataFrame.from_records(clique2df, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
           
            edges = []
            for index, rows in df.iterrows():
                lists = rows['entry1'] + rows['entry2']
                # Cliques here inherit neighbor weights, but will be overwritten by above
                cliques = [x for x in combinations(lists, 2)]
                network = [tup + (rows['types'], rows['value'], rows['name']) for tup in cliques]
                edges.append(network)
            # Standard line for obtaining a dataframe of edges
            edges2df = [e for edge in edges for e in edge]
            # Create pandas DF from edges
            df_out = pd.DataFrame.from_records(edges2df, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
            
            dft = df_out.groupby(['entry1', 'entry2'])['type'].apply(list).reset_index()
            dfv = df_out.groupby(['entry1', 'entry2'])['value'].apply(list).reset_index()
            dfn = df_out.groupby(['entry1', 'entry2'])['name'].apply(list).reset_index()
            dfx = dft
            dfx['value'] = dfv['value'].agg(','.join)
            dfx['name'] = dfn['name'].agg(','.join)
            dfx['type'] = dft['type'].agg(','.join)
            # Ensures independently parsed cliques overwrite the cliques, which inherited neighbor weights
            xdf = pd.concat([dfx, cliquedf]).drop_duplicates(subset = ['entry1', 'entry2'], keep = 'last')
            xdf.to_csv(wd / '{}.tsv'.format(pathway), sep = '\t', index = False)

def mixed_folder(input_data: str, wd: Path, unique: bool = False, graphics: bool = False):
    for file in Path(input_data).glob('*.xml'):
        tree = ET.parse(file)
        root = tree.getroot()
   
        title = root.get('title')
        pathway = root.get('name').replace('path:', '')
        species = root.get('org')
        pathway_link = root.get('link')
        if graphics:
            # Graphics
            graphics = graphics_dict(root)
            
            output_path = wd / 'kegg_mixed_network_{}'.format(species)
            output_path.mkdir(exist_ok = True)
            
            typer.echo(f'Now parsing: {title}...')
           
            e1 = []
            e2 = []
            n = []
            v = []
            for relation in root.findall('relation'):
                e1.append(relation.get('entry1'))
                e2.append(relation.get('entry2'))
                for subtype in relation.findall('subtype'):
                    n.append(subtype.get('name'))
                    v.append(subtype.get('value'))
           
            d = []
            for relation in root.findall('relation'):
                for subtype in relation:
                    d1=relation.attrib
                    d2=subtype.attrib
                    d3=json.dumps(d1),json.dumps(d2)
                    d.append(d3)    

            edgelist=[]
            # Takes json dictionary and removes formatting in order to obtain an edgelist
            # This odd line of code is necessary  because of the various nested and non-nested subtypes
            # This allows for edges to inherit weights from a nested subtype
            for line in d:
                x=line[0].replace("{","").replace("}","").replace('"entry1":',"")\
                .replace('"entry2":',"").replace('"type":',"").replace('"name":',"")\
                .replace('"value":',"").replace('"','').replace(',',"\t").replace(' ', '')
                y=line[1].replace("{","").replace("}","").replace('"name":',"")\
                .replace('"',"").replace('value:',"").replace(",",'\t').replace(' ', '')
                edgelist.append(x+"\t"+y)

            df=pd.DataFrame(edgelist)
       
            if df.empty:
                typer.echo(f'File "{file}" cannot be parsed.\nVisit {pathway_link} for pathway details.\nThere are likely no edges in which to parse...')
            else:
                df=df[0].str.split("\t", expand=True).rename({0: 'entry1',1: 'entry2',
                                                      2: 'types', 3:'name',
                                                      4: 'value'}, axis='columns')

                entry_dict = defaultdict(list)
                for entries in root.findall('entry'):
                    for key, items in entries.attrib.items():
                        entry_dict[key].append(items)  
                
                # Graphics
                df['pos1'] = df['entry1'].map(graphics)
                df['pos2'] = df['entry2'].map(graphics)
           
                entry_id=[]
                entry_name=[]
                entry_type=[]
                for key, items in entry_dict.items():
                    if key == 'id':
                        for i in items:
                            entry_id.append(i)
                    if key == 'name':
                        for i in items:
                            entry_name.append(i)
                    if key == 'type':
                        for i in items:
                            entry_type.append(i)
                
                if unique:
                    conversion_dictionary = conv_dict_unique(root)
                else:
                    conversion_dictionary = conv_dict(root)
           
                df['entry1'] = df['entry1'].map(conversion_dictionary)
                df['entry2'] = df['entry2'].map(conversion_dictionary)
                df['entry1'] = df['entry1'].astype(str).str.split(' ', expand = False)
                df['entry2'] = df['entry2'].astype(str).str.split(' ', expand = False)
           
                # Parse the cliques seperate so they won't inherit neighbor weights
                # Allows for custom weights so user knows which are customly parsed
                clique_edges = []
                for index, rows in df.iterrows():
                    if len(rows['entry1']) > 1:
                        cliques = [x for x in combinations(rows['entry1'], 2)]
                        cliques1 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                        clique_edges.append(cliques1)
                    if len(rows['entry2']) > 1:
                        cliques = [x for x in combinations(rows['entry2'], 2)]
                        cliques2 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                        clique_edges.append(cliques2)
                clique2df = [e for edge in clique_edges for e in edge]
                cliquedf = pd.DataFrame.from_records(clique2df, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
           
                edges = []
                for index, rows in df.iterrows():
                    lists = rows['entry1'] + rows['entry2']
                    # Cliques here inherit neighbor weights, but will be overwritten by above
                    cliques = [x for x in combinations(lists, 2)]
                    network = [tup + (rows['types'], rows['value'], rows['name'], rows['pos1'], rows['pos2']) for tup in cliques]
                    edges.append(network)
            
                # This removes edges which contain +p (phosphorylation) these oftentimes
                # overwrite important weight attributes while providing no vital information
                # edges2df = [e for edge in edges for e in edge if '+p' not in e and '-p' not in e]
            
                # Standard line for obtaining a dataframe of edges
                edges2df = [e for edge in edges for e in edge]
                # Create pandas DF from edges
                df_out = pd.DataFrame.from_records(edges2df, columns = ['entry1', 'entry2', 'type', 'value', 'name', 'pos1', 'pos2'])
                
                # Graphics
                pos_dict1 = {}
                pos_dict2 = {}
                for index, rows in df_out.iterrows():
                    pos_dict1[rows['entry1']] = rows['pos1']
                    pos_dict2[rows['entry2']] = rows['pos2']
                pos = pos_dict1 | pos_dict2
                json_dict = json.dumps(pos)
                with open(output_path / '{}_graphics.txt'.format(pathway), 'w') as outfile:
                    outfile.write(json_dict)
                
                
                dft = df_out.groupby(['entry1', 'entry2'])['type'].apply(list).reset_index()
                dfv = df_out.groupby(['entry1', 'entry2'])['value'].apply(list).reset_index()
                dfn = df_out.groupby(['entry1', 'entry2'])['name'].apply(list).reset_index()
                dfx = dft
                dfx['value'] = dfv['value'].agg(','.join)
                dfx['name'] = dfn['name'].agg(','.join)
                dfx['type'] = dft['type'].agg(','.join)
                # Ensures independently parsed cliques overwrite the cliques, which inherited neighbor weights
                xdf = pd.concat([dfx, cliquedf]).drop_duplicates(subset = ['entry1', 'entry2'], keep = 'last')
                xdf.to_csv(output_path / '{}.tsv'.format(pathway), sep = '\t', index = False)
        else:
            output_path = wd / 'kegg_mixed_network_{}'.format(species)
            output_path.mkdir(exist_ok = True)
            
            typer.echo(f'Now parsing: {title}...')
           
            e1 = []
            e2 = []
            n = []
            v = []
            for relation in root.findall('relation'):
                e1.append(relation.get('entry1'))
                e2.append(relation.get('entry2'))
                for subtype in relation.findall('subtype'):
                    n.append(subtype.get('name'))
                    v.append(subtype.get('value'))
           
            d = []
            for relation in root.findall('relation'):
                for subtype in relation:
                    d1=relation.attrib
                    d2=subtype.attrib
                    d3=json.dumps(d1),json.dumps(d2)
                    d.append(d3)    

            edgelist=[]
            # Takes json dictionary and removes formatting in order to obtain an edgelist
            # This odd line of code is necessary  because of the various nested and non-nested subtypes
            # This allows for edges to inherit weights from a nested subtype
            for line in d:
                x=line[0].replace("{","").replace("}","").replace('"entry1":',"")\
                .replace('"entry2":',"").replace('"type":',"").replace('"name":',"")\
                .replace('"value":',"").replace('"','').replace(',',"\t").replace(' ', '')
                y=line[1].replace("{","").replace("}","").replace('"name":',"")\
                .replace('"',"").replace('value:',"").replace(",",'\t').replace(' ', '')
                edgelist.append(x+"\t"+y)

            df=pd.DataFrame(edgelist)
       
            if df.empty:
                typer.echo(f'File "{file}" cannot be parsed.\nVisit {pathway_link} for pathway details.\nThere are likely no edges in which to parse...')
            else:
                df=df[0].str.split("\t", expand=True).rename({0: 'entry1',1: 'entry2',
                                                      2: 'types', 3:'name',
                                                      4: 'value'}, axis='columns')

                entry_dict = defaultdict(list)
                for entries in root.findall('entry'):
                    for key, items in entries.attrib.items():
                        entry_dict[key].append(items)
           
                entry_id=[]
                entry_name=[]
                entry_type=[]
                for key, items in entry_dict.items():
                    if key == 'id':
                        for i in items:
                            entry_id.append(i)
                    if key == 'name':
                        for i in items:
                            entry_name.append(i)
                    if key == 'type':
                        for i in items:
                            entry_type.append(i)
                
                if unique:
                    conversion_dictionary = conv_dict_unique(root)
                else:
                    conversion_dictionary = conv_dict(root)
           
                df['entry1'] = df['entry1'].map(conversion_dictionary)
                df['entry2'] = df['entry2'].map(conversion_dictionary)
                df['entry1'] = df['entry1'].astype(str).str.split(' ', expand = False)
                df['entry2'] = df['entry2'].astype(str).str.split(' ', expand = False)
           
                # Parse the cliques seperate so they won't inherit neighbor weights
                # Allows for custom weights so user knows which are customly parsed
                clique_edges = []
                for index, rows in df.iterrows():
                    if len(rows['entry1']) > 1:
                        cliques = [x for x in combinations(rows['entry1'], 2)]
                        cliques1 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                        clique_edges.append(cliques1)
                    if len(rows['entry2']) > 1:
                        cliques = [x for x in combinations(rows['entry2'], 2)]
                        cliques2 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                        clique_edges.append(cliques2)
                clique2df = [e for edge in clique_edges for e in edge]
                cliquedf = pd.DataFrame.from_records(clique2df, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
           
                edges = []
                for index, rows in df.iterrows():
                    lists = rows['entry1'] + rows['entry2']
                    # Cliques here inherit neighbor weights, but will be overwritten by above
                    cliques = [x for x in combinations(lists, 2)]
                    network = [tup + (rows['types'], rows['value'], rows['name']) for tup in cliques]
                    edges.append(network)
            
                # This removes edges which contain +p (phosphorylation) these oftentimes
                # overwrite important weight attributes while providing no vital information
                # edges2df = [e for edge in edges for e in edge if '+p' not in e and '-p' not in e]
            
                # Standard line for obtaining a dataframe of edges
                edges2df = [e for edge in edges for e in edge]
                # Create pandas DF from edges
                df_out = pd.DataFrame.from_records(edges2df, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
                dft = df_out.groupby(['entry1', 'entry2'])['type'].apply(list).reset_index()
                dfv = df_out.groupby(['entry1', 'entry2'])['value'].apply(list).reset_index()
                dfn = df_out.groupby(['entry1', 'entry2'])['name'].apply(list).reset_index()
                dfx = dft
                dfx['value'] = dfv['value'].agg(','.join)
                dfx['name'] = dfn['name'].agg(','.join)
                dfx['type'] = dft['type'].agg(','.join)
                # Ensures independently parsed cliques overwrite the cliques, which inherited neighbor weights
                xdf = pd.concat([dfx, cliquedf]).drop_duplicates(subset = ['entry1', 'entry2'], keep = 'last')
                xdf.to_csv(output_path / '{}.tsv'.format(pathway), sep = '\t', index = False)










#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Everest Uriel Castaneda
@desc: File for obtaining gene-only pathways
"""

import json
import typer
import pandas as pd
import networkx as nx
from pathlib import Path
from itertools import combinations
import xml.etree.ElementTree as ET
from collections import defaultdict

def conv_dict(root):
    # This dictionary is the default option
    # Only compounds are unique
    entry_dict = defaultdict(list)
    for entries in root.findall('entry'):
        for key, items in entries.attrib.items():
            entry_dict[key].append(items)  
       
    entry_id=[]
    entry_name=[]
    for key, items in entry_dict.items():
        if key == 'id':
            for i in items:
                entry_id.append(i)
        if key == 'name':
            for i in items:
                entry_name.append(i)

    unique_compound = []
    for i in range(0, len(entry_id)):
        c = [name + '-' + entry_id[i] if name.startswith('cpd:') or name == 'undefined' else name for name in entry_name[i].split()]
        unique_compound.append(' '.join(c))

    conversion_dictionary = dict(zip(entry_id, unique_compound))
    return conversion_dictionary

def conv_dict_unique(root):
    # This dictionary is the unique version
    # Every item is unique to reveal subgraphs
    entry_dict = defaultdict(list)
    for entries in root.findall('entry'):
        for key, items in entries.attrib.items():
            entry_dict[key].append(items)  
       
    entry_id=[]
    entry_name=[]
    for key, items in entry_dict.items():
        if key == 'id':
            for i in items:
                entry_id.append(i)
        if key == 'name':
            for i in items:
                entry_name.append(i)
   
    unique_name = []
    for i in range(0, len(entry_id)):
        e = [name + '-' + entry_id[i] for name in entry_name[i].split()]
        unique_name.append(' '.join(e))

    conversion_dictionary = dict(zip(entry_id, unique_name))
    return conversion_dictionary

def graphics_dict(root):
    entry_dict = defaultdict(list)
    for entry in root.findall('entry'):
        for graphics in entry.find('graphics').items():
            if graphics[0] == 'x' or graphics[0] == 'y':
                entry_dict[entry.attrib['id']].append(int(graphics[1]))
    graphics_dict = {}
    for key, items in entry_dict.items():
        graphics_dict[key] = tuple(items)
    return graphics_dict

def genes_file(input_data: str, wd, unique: bool = False, graphics: bool = False):
    """
    Converts a folder of KGML files or a single KGML file into a weighted
    edgelist of genes that can be used in graph analysis. If -u/--unique flag 
    is used genes are returned with terminal modifiers to enhance network 
    visualization or analysis. If the -g/--graphics flag is used, x-y 
    coordinates of pathways are returned, which may be used as positions in 
    NetworkX\'s graph drawing commands.
    """
    tree = ET.parse(input_data)
    root = tree.getroot()
   
    title = root.get('title')
    pathway = root.get('name').replace('path:', '')
    pathway_link = root.get('link')
    if graphics:
        # Graphics
        graphics = graphics_dict(root)
        
        typer.echo(f'Now parsing: {title}...')
           
        e1 = []
        e2 = []
        n = []
        v = []
        for relation in root.findall('relation'):
            e1.append(relation.get('entry1'))
            e2.append(relation.get('entry2'))
            for subtype in relation.findall('subtype'):
                n.append(subtype.get('name'))
                v.append(subtype.get('value'))
           
        d = []
        for relation in root.findall('relation'):
            for subtype in relation:
                d1=relation.attrib
                d2=subtype.attrib
                d3=json.dumps(d1),json.dumps(d2)
                d.append(d3)    

        edgelist=[]
        # Takes json dictionary and removes formatting in order to obtain an edgelist
        # This odd line of code is necessary  because of the various nested and non-nested subtypes
        # This allows for edges to inherit weights from a nested subtype
        for line in d:
            x=line[0].replace("{","").replace("}","").replace('"entry1":',"")\
                .replace('"entry2":',"").replace('"type":',"").replace('"name":',"")\
                .replace('"value":',"").replace('"','').replace(',',"\t").replace(' ', '')
            y=line[1].replace("{","").replace("}","").replace('"name":',"")\
                .replace('"',"").replace('value:',"").replace(",",'\t').replace(' ', '')
            edgelist.append(x+"\t"+y)

        df=pd.DataFrame(edgelist)
       
        if df.empty:
            typer.echo(f'File "{input_data}" cannot be parsed.\nVisit {pathway_link} for pathway details.\nThere are likely no edges in which to parse...')
        else:
            df=df[0].str.split("\t", expand=True).rename({0: 'entry1',1: 'entry2',
                                                      2: 'types', 3:'name',
                                                      4: 'value'}, axis='columns')

            entry_dict = defaultdict(list)
            for entries in root.findall('entry'):
                for key, items in entries.attrib.items():
                    entry_dict[key].append(items)  
                    
            # Graphics
            df['pos1'] = df['entry1'].map(graphics)
            df['pos2'] = df['entry2'].map(graphics)
           
            entry_id=[]
            entry_name=[]
            entry_type=[]
            for key, items in entry_dict.items():
                if key == 'id':
                    for i in items:
                        entry_id.append(i)
                if key == 'name':
                    for i in items:
                        entry_name.append(i)
                if key == 'type':
                    for i in items:
                        entry_type.append(i)
            
            if unique:
                conversion_dictionary = conv_dict_unique(root)
            else:
                conversion_dictionary = conv_dict(root)
           
            df['entry1'] = df['entry1'].map(conversion_dictionary)
            df['entry2'] = df['entry2'].map(conversion_dictionary)
            df['entry1'] = df['entry1'].astype(str).str.split(' ', expand = False)
            df['entry2'] = df['entry2'].astype(str).str.split(' ', expand = False)
           
            # Parse the cliques seperate so they won't inherit neighbor weights
            # Allows for custom weights so user knows which are customly parsed
            clique_edges = []
            for index, rows in df.iterrows():
                if len(rows['entry1']) > 1:
                    cliques = [x for x in combinations(rows['entry1'], 2)]
                    cliques1 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                    clique_edges.append(cliques1)
                if len(rows['entry2']) > 1:
                    cliques = [x for x in combinations(rows['entry2'], 2)]
                    cliques2 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                    clique_edges.append(cliques2)
            clique2df = [e for edge in clique_edges for e in edge]
            cliquedf = pd.DataFrame.from_records(clique2df, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
           
            edges = []
            for index, rows in df.iterrows():
                lists = rows['entry1'] + rows['entry2']
                # Cliques here inherit neighbor weights, but will be overwritten by above
                cliques = [x for x in combinations(lists, 2)]
                network = [tup + (rows['types'], rows['value'], rows['name'], rows['pos1'], rows['pos2']) for tup in cliques]
                edges.append(network)
            
            # This removes edges which contain +p (phosphorylation) these oftentimes
            # overwrite important weight attributes while providing no vital information
            # edges2df = [e for edge in edges for e in edge if '+p' not in e and '-p' not in e]
            
            # Standard line for obtaining a dataframe of edges
            edges2df = [e for edge in edges for e in edge]
            # Create pandas DF from edges
            df_out = pd.DataFrame.from_records(edges2df, columns = ['entry1', 'entry2', 'type', 'value', 'name', 'pos1', 'pos2'])
            
            # Graphics
            pos_dict1 = {}
            pos_dict2 = {}
            for index, rows in df_out.iterrows():
                pos_dict1[rows['entry1']] = rows['pos1']
                pos_dict2[rows['entry2']] = rows['pos2']
            pos = pos_dict1 | pos_dict2
            json_dict = json.dumps(pos)
            with open('%s_graphics.txt' % pathway, 'w') as outfile:
                outfile.write(json_dict)
            
            dft = df_out.groupby(['entry1', 'entry2'])['type'].apply(list).reset_index()
            dfv = df_out.groupby(['entry1', 'entry2'])['value'].apply(list).reset_index()
            dfn = df_out.groupby(['entry1', 'entry2'])['name'].apply(list).reset_index()
            dfx = dft
            dfx['value'] = dfv['value'].agg(','.join)
            dfx['name'] = dfn['name'].agg(','.join)
            dfx['type'] = dft['type'].agg(','.join)
            # Ensures independently parsed cliques overwrite the cliques, which inherited neighbor weights
            xdf = pd.concat([dfx, cliquedf]).drop_duplicates(subset = ['entry1', 'entry2'], keep = 'last')

            if xdf[(xdf['entry1'].str.startswith('cpd:')) | (xdf['entry2'].str.startswith('cpd:'))].empty == True and xdf[(xdf['entry1'].str.startswith('undefined')) | (xdf['entry2'].str.startswith('undefined'))].empty == True:
                # If pathway has no compounds or undefined nodes, write file
                xdf_out = xdf[(~xdf['entry1'].str.startswith('path')) & (~xdf['entry2'].str.startswith('path'))]
                xdf_out.to_csv(wd / '{}.tsv'.format(pathway), sep = '\t', index = False)
            else:
                # These next series of steps are for propagating compounds and undefined nodes
                # Uses networkx
                G = nx.from_pandas_edgelist(xdf, source = 'entry1', target = 'entry2', edge_attr = 'name', create_using = nx.DiGraph())
                new_edges = []
                for node in G.nodes:
                    # Gather roots and leaflets
                    roots = [n for n, d in G.in_degree() if d == 0]
                    leaflets = [n for n, d in G.out_degree() if d == 0]
                    # Find compounds or undefined proteins that might need propagation
                    if node.startswith('cpd') or node.startswith('undefined'):
                        if node in roots or node in leaflets:
                            # Passes root nodes and leaflet nodes
                            # These do not need propagation as they terminate
                            pass
                        else:
                            out_edges = [e for e in G.out_edges(node)]
                            in_edges = [e for e in G.in_edges(node)]
                            for i in in_edges:
                                for o in out_edges:
                                    if not i[0].startswith('cpd') and not o[1].startswith('cpd') and not i[0].startswith('undefined') and not o[1].startswith('undefined') and not i[0].startswith('path') and not o[1].startswith('path'):
                                        # Simple compound propagation removes compound between two genes
                                        # Example: hsa:xxx -> cpd:xxx -> hsa:xxx to hsa:xxx -> hsa:xxx
                                        new_edges.append([i[0], o[1], 'CPp', 'Custom', 'compound propagation'])
                                    else:
                                        # Loops through roots and leaves to find the start and end of each node that is propagated
                                        # These pathways have copious amounts of compounds and undefined genes in between the propagated genes
                                        # This loop also picks up on non-proteins that are left out if the above simple compound propagation is used
                                        for root in roots:
                                            for leaf in leaflets:
                                                if nx.has_path(G, root, node) == True and nx.has_path(G, node, leaf) == True:
                                                    # Uses the shortest path between the root/leaf and query node
                                                    rpath = [p for p in nx.shortest_path(G, root, node)]
                                                    lpath = [p for p in nx.shortest_path(G, node, leaf)]
                                                    # Skip paths that have no proteins
                                                    if all(node.startswith('cpd') or node.startswith('undefined') or node.startswith('path') for node in rpath) == True or all(node.startswith('cpd') or node.startswith('undefined') or node.startswith('path') for node in lpath) == True:
                                                        pass
                                                    else:
                                                        # Indexes all elements in the shortest path list that is a gene
                                                        rindex = [index for index, element in enumerate(rpath) if not element.startswith('cpd') and not element.startswith('undefined') and not element.startswith('path')]
                                                        lindex = [index for index, element in enumerate(lpath) if not element.startswith('cpd') and not element.startswith('undefined') and not element.startswith('path')]
                                                        # Here we use the maximum of the root path and the minimum of the leaf path to successfully propagate compounds
                                                        # Note below example
                                                        # [mmu:x, cpd:###, cpd:x] --> max is mmu:x 
                                                        # [cpd:x, cpd:###, undefined, mmu:y, mmu:###] --> min is mmu:y
                                                        # creates edge [mmu:x, mmu:y]
                                                        new_edges.append([rpath[max(rindex)], lpath[min(lindex)], 'CPp', 'Custom', 'compound propagation'])
                # Creates a dataframe of the new edges that are a result of compound propagation
                new_edges_df = pd.DataFrame(new_edges, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
                # Concatenates the new edges with the edges from the above (cliques and original parsed edges)
                df0 = pd.concat([xdf, new_edges_df])
                # Drop any duplicated edges
                df1 = df0.drop_duplicates()
                # Removes compounds and undefined as they were propagated and no longer needed
                df2 = df1[(~df1['entry1'].str.startswith('cpd')) & (~df1['entry2'].str.startswith('cpd')) & (~df1['entry1'].str.startswith('undefined')) & (~df1['entry2'].str.startswith('undefined')) & (~df1['entry1'].str.startswith('path')) & (~df1['entry2'].str.startswith('path'))]
                df2.to_csv(wd / '{}.tsv'.format(pathway), sep = '\t', index = False)
    else:
        typer.echo(f'Now parsing: {title}...')
           
        e1 = []
        e2 = []
        n = []
        v = []
        for relation in root.findall('relation'):
            e1.append(relation.get('entry1'))
            e2.append(relation.get('entry2'))
            for subtype in relation.findall('subtype'):
                n.append(subtype.get('name'))
                v.append(subtype.get('value'))
           
        d = []
        for relation in root.findall('relation'):
            for subtype in relation:
                d1=relation.attrib
                d2=subtype.attrib
                d3=json.dumps(d1),json.dumps(d2)
                d.append(d3)    

        edgelist=[]
        # Takes json dictionary and removes formatting in order to obtain an edgelist
        # This odd line of code is necessary  because of the various nested and non-nested subtypes
        # This allows for edges to inherit weights from a nested subtype
        for line in d:
            x=line[0].replace("{","").replace("}","").replace('"entry1":',"")\
                .replace('"entry2":',"").replace('"type":',"").replace('"name":',"")\
                .replace('"value":',"").replace('"','').replace(',',"\t").replace(' ', '')
            y=line[1].replace("{","").replace("}","").replace('"name":',"")\
                .replace('"',"").replace('value:',"").replace(",",'\t').replace(' ', '')
            edgelist.append(x+"\t"+y)

        df=pd.DataFrame(edgelist)
       
        if df.empty:
            typer.echo(f'File "{input_data}" cannot be parsed.\nVisit {pathway_link} for pathway details.\nThere are likely no edges in which to parse...')
        else:
            df=df[0].str.split("\t", expand=True).rename({0: 'entry1',1: 'entry2',
                                                      2: 'types', 3:'name',
                                                      4: 'value'}, axis='columns')

            entry_dict = defaultdict(list)
            for entries in root.findall('entry'):
                for key, items in entries.attrib.items():
                    entry_dict[key].append(items)
           
            entry_id=[]
            entry_name=[]
            entry_type=[]
            for key, items in entry_dict.items():
                if key == 'id':
                    for i in items:
                        entry_id.append(i)
                if key == 'name':
                    for i in items:
                        entry_name.append(i)
                if key == 'type':
                    for i in items:
                        entry_type.append(i)
            
            if unique:
                conversion_dictionary = conv_dict_unique(root)
            else:
                conversion_dictionary = conv_dict(root)
           
            df['entry1'] = df['entry1'].map(conversion_dictionary)
            df['entry2'] = df['entry2'].map(conversion_dictionary)
            df['entry1'] = df['entry1'].astype(str).str.split(' ', expand = False)
            df['entry2'] = df['entry2'].astype(str).str.split(' ', expand = False)
           
            # Parse the cliques seperate so they won't inherit neighbor weights
            # Allows for custom weights so user knows which are customly parsed
            clique_edges = []
            for index, rows in df.iterrows():
                if len(rows['entry1']) > 1:
                    cliques = [x for x in combinations(rows['entry1'], 2)]
                    cliques1 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                    clique_edges.append(cliques1)
                if len(rows['entry2']) > 1:
                    cliques = [x for x in combinations(rows['entry2'], 2)]
                    cliques2 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                    clique_edges.append(cliques2)
            clique2df = [e for edge in clique_edges for e in edge]
            cliquedf = pd.DataFrame.from_records(clique2df, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
           
            edges = []
            for index, rows in df.iterrows():
                lists = rows['entry1'] + rows['entry2']
                # Cliques here inherit neighbor weights, but will be overwritten by above
                cliques = [x for x in combinations(lists, 2)]
                network = [tup + (rows['types'], rows['value'], rows['name']) for tup in cliques]
                edges.append(network)
            
            # This removes edges which contain +p (phosphorylation) these oftentimes
            # overwrite important weight attributes while providing no vital information
            # edges2df = [e for edge in edges for e in edge if '+p' not in e and '-p' not in e]
            
            # Standard line for obtaining a dataframe of edges
            edges2df = [e for edge in edges for e in edge]
            # Create pandas DF from edges
            df_out = pd.DataFrame.from_records(edges2df, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
            
            dft = df_out.groupby(['entry1', 'entry2'])['type'].apply(list).reset_index()
            dfv = df_out.groupby(['entry1', 'entry2'])['value'].apply(list).reset_index()
            dfn = df_out.groupby(['entry1', 'entry2'])['name'].apply(list).reset_index()
            dfx = dft
            dfx['value'] = dfv['value'].agg(','.join)
            dfx['name'] = dfn['name'].agg(','.join)
            dfx['type'] = dft['type'].agg(','.join)
            # Ensures independently parsed cliques overwrite the cliques, which inherited neighbor weights
            xdf = pd.concat([dfx, cliquedf]).drop_duplicates(subset = ['entry1', 'entry2'], keep = 'last')

            if xdf[(xdf['entry1'].str.startswith('cpd:')) | (xdf['entry2'].str.startswith('cpd:'))].empty == True and xdf[(xdf['entry1'].str.startswith('undefined')) | (xdf['entry2'].str.startswith('undefined'))].empty == True:
                # If pathway has no compounds or undefined nodes, write file
                xdf_out = xdf[(~xdf['entry1'].str.startswith('path')) & (~xdf['entry2'].str.startswith('path'))]
                xdf_out.to_csv(wd / '{}.tsv'.format(pathway), sep = '\t', index = False)
            else:
                # These next series of steps are for propagating compounds and undefined nodes
                # Uses networkx
                G = nx.from_pandas_edgelist(xdf, source = 'entry1', target = 'entry2', edge_attr = 'name', create_using = nx.DiGraph())
                new_edges = []
                for node in G.nodes:
                    # Gather roots and leaflets
                    roots = [n for n, d in G.in_degree() if d == 0]
                    leaflets = [n for n, d in G.out_degree() if d == 0]
                    # Find compounds or undefined proteins that might need propagation
                    if node.startswith('cpd') or node.startswith('undefined'):
                        if node in roots or node in leaflets:
                            # Passes root nodes and leaflet nodes
                            # These do not need propagation as they terminate
                            pass
                        else:
                            out_edges = [e for e in G.out_edges(node)]
                            in_edges = [e for e in G.in_edges(node)]
                            for i in in_edges:
                                for o in out_edges:
                                    if not i[0].startswith('cpd') and not o[1].startswith('cpd') and not i[0].startswith('undefined') and not o[1].startswith('undefined') and not i[0].startswith('path') and not o[1].startswith('path'):
                                        # Simple compound propagation removes compound between two genes
                                        # Example: hsa:xxx -> cpd:xxx -> hsa:xxx to hsa:xxx -> hsa:xxx
                                        new_edges.append([i[0], o[1], 'CPp', 'Custom', 'compound propagation'])
                                    else:
                                        # Loops through roots and leaves to find the start and end of each node that is propagated
                                        # These pathways have copious amounts of compounds and undefined genes in between the propagated genes
                                        # This loop also picks up on non-proteins that are left out if the above simple compound propagation is used
                                        for root in roots:
                                            for leaf in leaflets:
                                                if nx.has_path(G, root, node) == True and nx.has_path(G, node, leaf) == True:
                                                    # Uses the shortest path between the root/leaf and query node
                                                    rpath = [p for p in nx.shortest_path(G, root, node)]
                                                    lpath = [p for p in nx.shortest_path(G, node, leaf)]
                                                    # Skip paths that have no proteins
                                                    if all(node.startswith('cpd') or node.startswith('undefined') or node.startswith('path') for node in rpath) == True or all(node.startswith('cpd') or node.startswith('undefined') or node.startswith('path') for node in lpath) == True:
                                                        pass
                                                    else:
                                                        # Indexes all elements in the shortest path list that is a gene
                                                        rindex = [index for index, element in enumerate(rpath) if not element.startswith('cpd') and not element.startswith('undefined') and not element.startswith('path')]
                                                        lindex = [index for index, element in enumerate(lpath) if not element.startswith('cpd') and not element.startswith('undefined') and not element.startswith('path')]
                                                        # Here we use the maximum of the root path and the minimum of the leaf path to successfully propagate compounds
                                                        # Note below example
                                                        # [mmu:x, cpd:###, cpd:x] --> max is mmu:x 
                                                        # [cpd:x, cpd:###, undefined, mmu:y, mmu:###] --> min is mmu:y
                                                        # creates edge [mmu:x, mmu:y]
                                                        new_edges.append([rpath[max(rindex)], lpath[min(lindex)], 'CPp', 'Custom', 'compound propagation'])
                # Creates a dataframe of the new edges that are a result of compound propagation
                new_edges_df = pd.DataFrame(new_edges, columns = df_out.columns)
                # Concatenates the new edges with the edges from the above (cliques and original parsed edges)
                df0 = pd.concat([xdf, new_edges_df])
                # Drop any duplicated edges
                df1 = df0.drop_duplicates()
                # Removes compounds and undefined as they were propagated and no longer needed
                df2 = df1[(~df1['entry1'].str.startswith('cpd')) & (~df1['entry2'].str.startswith('cpd')) & (~df1['entry1'].str.startswith('undefined')) & (~df1['entry2'].str.startswith('undefined')) & (~df1['entry1'].str.startswith('path')) & (~df1['entry2'].str.startswith('path'))]
                df2.to_csv(wd / '{}.tsv'.format(pathway), sep = '\t', index = False)

def genes_folder(input_data: str, wd: Path, unique: bool = False, graphics: bool = False):
    for file in Path(input_data).glob('*.xml'):
        tree = ET.parse(file)
        root = tree.getroot()
   
        title = root.get('title')
        pathway = root.get('name').replace('path:', '')
        species = root.get('org')
        pathway_link = root.get('link')
        if graphics:
            # Graphics
            graphics = graphics_dict(root)
            
            output_path = wd / 'kegg_gene_network_{}'.format(species)
            output_path.mkdir(exist_ok = True)
            
            typer.echo(f'Now parsing: {title}...')
            e1 = []
            e2 = []
            n = []
            v = []
            for relation in root.findall('relation'):
                e1.append(relation.get('entry1'))
                e2.append(relation.get('entry2'))
                for subtype in relation.findall('subtype'):
                    n.append(subtype.get('name'))
                    v.append(subtype.get('value'))
           
            d = []
            for relation in root.findall('relation'):
                for subtype in relation:
                    d1=relation.attrib
                    d2=subtype.attrib
                    d3=json.dumps(d1),json.dumps(d2)
                    d.append(d3)    

            edgelist=[]
            # Takes json dictionary and removes formatting in order to obtain an edgelist
            # This odd line of code is necessary  because of the various nested and non-nested subtypes
            # This allows for edges to inherit weights from a nested subtype
            for line in d:
                x=line[0].replace("{","").replace("}","").replace('"entry1":',"")\
                .replace('"entry2":',"").replace('"type":',"").replace('"name":',"")\
                .replace('"value":',"").replace('"','').replace(',',"\t").replace(' ', '')
                y=line[1].replace("{","").replace("}","").replace('"name":',"")\
                .replace('"',"").replace('value:',"").replace(",",'\t').replace(' ', '')
                edgelist.append(x+"\t"+y)

            df=pd.DataFrame(edgelist)
       
            if df.empty:
                typer.echo(f'File "{file}" cannot be parsed.\nVisit {pathway_link} for pathway details.\nThere are likely no edges in which to parse...')
            else:
                df=df[0].str.split("\t", expand=True).rename({0: 'entry1',1: 'entry2',
                                                      2: 'types', 3:'name',
                                                      4: 'value'}, axis='columns')

                entry_dict = defaultdict(list)
                for entries in root.findall('entry'):
                    for key, items in entries.attrib.items():
                        entry_dict[key].append(items) 
                        
                # Graphics
                df['pos1'] = df['entry1'].map(graphics)
                df['pos2'] = df['entry2'].map(graphics)
           
                entry_id=[]
                entry_name=[]
                entry_type=[]
                for key, items in entry_dict.items():
                    if key == 'id':
                        for i in items:
                            entry_id.append(i)
                    if key == 'name':
                        for i in items:
                            entry_name.append(i)
                    if key == 'type':
                        for i in items:
                            entry_type.append(i)
                
                if unique:
                    conversion_dictionary = conv_dict_unique(root)
                else:
                    conversion_dictionary = conv_dict(root)
           
                df['entry1'] = df['entry1'].map(conversion_dictionary)
                df['entry2'] = df['entry2'].map(conversion_dictionary)
                df['entry1'] = df['entry1'].astype(str).str.split(' ', expand = False)
                df['entry2'] = df['entry2'].astype(str).str.split(' ', expand = False)
           
                # Parse the cliques seperate so they won't inherit neighbor weights
                # Allows for custom weights so user knows which are customly parsed
                clique_edges = []
                for index, rows in df.iterrows():
                    if len(rows['entry1']) > 1:
                        cliques = [x for x in combinations(rows['entry1'], 2)]
                        cliques1 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                        clique_edges.append(cliques1)
                    if len(rows['entry2']) > 1:
                        cliques = [x for x in combinations(rows['entry2'], 2)]
                        cliques2 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                        clique_edges.append(cliques2)
                clique2df = [e for edge in clique_edges for e in edge]
                cliquedf = pd.DataFrame.from_records(clique2df, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
           
                edges = []
                for index, rows in df.iterrows():
                    lists = rows['entry1'] + rows['entry2']
                    # Cliques here inherit neighbor weights, but will be overwritten by above
                    cliques = [x for x in combinations(lists, 2)]
                    network = [tup + (rows['types'], rows['value'], rows['name'], rows['pos1'], rows['pos2']) for tup in cliques]
                    edges.append(network)
            
                # This removes edges which contain +p (phosphorylation) these oftentimes
                # overwrite important weight attributes while providing no vital information
                # edges2df = [e for edge in edges for e in edge if '+p' not in e and '-p' not in e]
            
                # Standard line for obtaining a dataframe of edges
                edges2df = [e for edge in edges for e in edge]
                # Create pandas DF from edges
                df_out = pd.DataFrame.from_records(edges2df, columns = ['entry1', 'entry2', 'type', 'value', 'name', 'pos1', 'pos2'])
                
                # Graphics
                if graphics:
                    pos_dict1 = {}
                    pos_dict2 = {}
                    for index, rows in df_out.iterrows():
                        pos_dict1[rows['entry1']] = rows['pos1']
                        pos_dict2[rows['entry2']] = rows['pos2']
                    pos = pos_dict1 | pos_dict2
                    json_dict = json.dumps(pos)
                    with open(output_path / '{}_graphics.txt'.format(pathway), 'w') as outfile:
                        outfile.write(json_dict)
                
                dft = df_out.groupby(['entry1', 'entry2'])['type'].apply(list).reset_index()
                dfv = df_out.groupby(['entry1', 'entry2'])['value'].apply(list).reset_index()
                dfn = df_out.groupby(['entry1', 'entry2'])['name'].apply(list).reset_index()
                dfx = dft
                dfx['value'] = dfv['value'].agg(','.join)
                dfx['name'] = dfn['name'].agg(','.join)
                dfx['type'] = dft['type'].agg(','.join)
                # Ensures independently parsed cliques overwrite the cliques, which inherited neighbor weights
                xdf = pd.concat([dfx, cliquedf]).drop_duplicates(subset = ['entry1', 'entry2'], keep = 'last')

                if xdf[(xdf['entry1'].str.startswith('cpd:')) | (xdf['entry2'].str.startswith('cpd:'))].empty == True and xdf[(xdf['entry1'].str.startswith('undefined')) | (xdf['entry2'].str.startswith('undefined'))].empty == True:
                    # If pathway has no compounds or undefined nodes, write file
                    # Note: since this isn't a mixed pathway, "path:" nodes are excluded
                    xdf_out = xdf[(~xdf['entry1'].str.startswith('path')) & (~xdf['entry2'].str.startswith('path'))]
                    xdf_out.to_csv(output_path / '{}.tsv'.format(pathway), sep = '\t', index = False)
                else:
                    # These next series of steps are for propagating compounds and undefined nodes
                    # Uses networkx
                    G = nx.from_pandas_edgelist(xdf, source = 'entry1', target = 'entry2', edge_attr = 'name', create_using = nx.DiGraph())
                    new_edges = []
                    for node in G.nodes:
                        # Gather roots and leaflets
                        roots = [n for n, d in G.in_degree() if d == 0]
                        leaflets = [n for n, d in G.out_degree() if d == 0]
                        # Find compounds or undefined proteins that might need propagation
                        if node.startswith('cpd') or node.startswith('undefined'):
                            if node in roots or node in leaflets:
                                # Passes root nodes and leaflet nodes
                                # These do not need propagation as they terminate
                                pass
                            else:
                                out_edges = [e for e in G.out_edges(node)]
                                in_edges = [e for e in G.in_edges(node)]
                                for i in in_edges:
                                    for o in out_edges:
                                        if not i[0].startswith('cpd') and not o[1].startswith('cpd') and not i[0].startswith('undefined') and not o[1].startswith('undefined') and not i[0].startswith('path') and not o[1].startswith('path'):
                                            # Simple compound propagation removes compound between two genes
                                            # Example: hsa:xxx -> cpd:xxx -> hsa:xxx to hsa:xxx -> hsa:xxx
                                            new_edges.append([i[0], o[1], 'CPp', 'Custom', 'compound propagation'])
                                        else:
                                            # Loops through roots and leaves to find the start and end of each node that is propagated
                                            # These pathways have copious amounts of compounds and undefined genes in between the propagated genes
                                            # This loop also picks up on non-proteins that are left out if the above simple compound propagation is used
                                            for root in roots:
                                                for leaf in leaflets:
                                                    if nx.has_path(G, root, node) == True and nx.has_path(G, node, leaf) == True:
                                                        # Uses the shortest path between the root/leaf and query node
                                                        rpath = [p for p in nx.shortest_path(G, root, node)]
                                                        lpath = [p for p in nx.shortest_path(G, node, leaf)]
                                                        # Skip paths that have no proteins
                                                        if all(node.startswith('cpd') or node.startswith('undefined') or node.startswith('path') for node in rpath) == True or all(node.startswith('cpd') or node.startswith('undefined') or node.startswith('path') for node in lpath) == True:
                                                            pass
                                                        else:
                                                            # Indexes all elements in the shortest path list that is a gene
                                                            rindex = [index for index, element in enumerate(rpath) if not element.startswith('cpd') and not element.startswith('undefined') and not element.startswith('path')]
                                                            lindex = [index for index, element in enumerate(lpath) if not element.startswith('cpd') and not element.startswith('undefined') and not element.startswith('path')]
                                                            # Here we use the maximum of the root path and the minimum of the leaf path to successfully propagate compounds
                                                            # Note below example
                                                            # [mmu:x, cpd:###, cpd:x] --> max is mmu:x 
                                                            # [cpd:x, cpd:###, undefined, mmu:y, mmu:###] --> min is mmu:y
                                                            # creates edge [mmu:x, mmu:y]
                                                            new_edges.append([rpath[max(rindex)], lpath[min(lindex)], 'CPp', 'Custom', 'compound propagation'])
                    # Creates a dataframe of the new edges that are a result of compound propagation
                    new_edges_df = pd.DataFrame(new_edges, columns = df_out.drop(['pos1', 'pos2'], axis = 1).columns)
                    # Concatenates the new edges with the edges from the above (cliques and original parsed edges)
                    df0 = pd.concat([xdf, new_edges_df])
                    # Drop any duplicated edges
                    df1 = df0.drop_duplicates()
                    # Removes compounds and undefined as they were propagated and no longer needed
                    df2 = df1[(~df1['entry1'].str.startswith('cpd')) & (~df1['entry2'].str.startswith('cpd')) & (~df1['entry1'].str.startswith('undefined')) & (~df1['entry2'].str.startswith('undefined')) & (~df1['entry1'].str.startswith('path')) & (~df1['entry2'].str.startswith('path'))]
                    df2.to_csv(output_path / '{}.tsv'.format(pathway), sep = '\t', index = False)
        else:
            output_path = wd / 'kegg_gene_network_{}'.format(species)
            output_path.mkdir(exist_ok = True)
            
            typer.echo(f'Now parsing: {title}...')
           
            e1 = []
            e2 = []
            n = []
            v = []
            for relation in root.findall('relation'):
                e1.append(relation.get('entry1'))
                e2.append(relation.get('entry2'))
                for subtype in relation.findall('subtype'):
                    n.append(subtype.get('name'))
                    v.append(subtype.get('value'))
           
            d = []
            for relation in root.findall('relation'):
                for subtype in relation:
                    d1=relation.attrib
                    d2=subtype.attrib
                    d3=json.dumps(d1),json.dumps(d2)
                    d.append(d3)    

            edgelist=[]
            # Takes json dictionary and removes formatting in order to obtain an edgelist
            # This odd line of code is necessary  because of the various nested and non-nested subtypes
            # This allows for edges to inherit weights from a nested subtype
            for line in d:
                x=line[0].replace("{","").replace("}","").replace('"entry1":',"")\
                .replace('"entry2":',"").replace('"type":',"").replace('"name":',"")\
                .replace('"value":',"").replace('"','').replace(',',"\t").replace(' ', '')
                y=line[1].replace("{","").replace("}","").replace('"name":',"")\
                .replace('"',"").replace('value:',"").replace(",",'\t').replace(' ', '')
                edgelist.append(x+"\t"+y)

            df=pd.DataFrame(edgelist)
       
            if df.empty:
                typer.echo(f'File "{file}" cannot be parsed.\nVisit {pathway_link} for pathway details.\nThere are likely no edges in which to parse...')
            else:
                df=df[0].str.split("\t", expand=True).rename({0: 'entry1',1: 'entry2',
                                                      2: 'types', 3:'name',
                                                      4: 'value'}, axis='columns')

                entry_dict = defaultdict(list)
                for entries in root.findall('entry'):
                    for key, items in entries.attrib.items():
                        entry_dict[key].append(items)
           
                entry_id=[]
                entry_name=[]
                entry_type=[]
                for key, items in entry_dict.items():
                    if key == 'id':
                        for i in items:
                            entry_id.append(i)
                    if key == 'name':
                        for i in items:
                            entry_name.append(i)
                    if key == 'type':
                        for i in items:
                            entry_type.append(i)
                
                if unique:
                    conversion_dictionary = conv_dict_unique(root)
                else:
                    conversion_dictionary = conv_dict(root)
           
                df['entry1'] = df['entry1'].map(conversion_dictionary)
                df['entry2'] = df['entry2'].map(conversion_dictionary)
                df['entry1'] = df['entry1'].astype(str).str.split(' ', expand = False)
                df['entry2'] = df['entry2'].astype(str).str.split(' ', expand = False)
           
                # Parse the cliques seperate so they won't inherit neighbor weights
                # Allows for custom weights so user knows which are customly parsed
                clique_edges = []
                for index, rows in df.iterrows():
                    if len(rows['entry1']) > 1:
                        cliques = [x for x in combinations(rows['entry1'], 2)]
                        cliques1 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                        clique_edges.append(cliques1)
                    if len(rows['entry2']) > 1:
                        cliques = [x for x in combinations(rows['entry2'], 2)]
                        cliques2 = [tup + ('type 2', 'undirectional', 'clique') for tup in cliques]
                        clique_edges.append(cliques2)
                clique2df = [e for edge in clique_edges for e in edge]
                cliquedf = pd.DataFrame.from_records(clique2df, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
           
                edges = []
                for index, rows in df.iterrows():
                    lists = rows['entry1'] + rows['entry2']
                    # Cliques here inherit neighbor weights, but will be overwritten by above
                    cliques = [x for x in combinations(lists, 2)]
                    network = [tup + (rows['types'], rows['value'], rows['name']) for tup in cliques]
                    edges.append(network)
            
                # This removes edges which contain +p (phosphorylation) these oftentimes
                # overwrite important weight attributes while providing no vital information
                # edges2df = [e for edge in edges for e in edge if '+p' not in e and '-p' not in e]
            
                # Standard line for obtaining a dataframe of edges
                edges2df = [e for edge in edges for e in edge]
                # Create pandas DF from edges
                df_out = pd.DataFrame.from_records(edges2df, columns = ['entry1', 'entry2', 'type', 'value', 'name'])
                
                dft = df_out.groupby(['entry1', 'entry2'])['type'].apply(list).reset_index()
                dfv = df_out.groupby(['entry1', 'entry2'])['value'].apply(list).reset_index()
                dfn = df_out.groupby(['entry1', 'entry2'])['name'].apply(list).reset_index()
                dfx = dft
                dfx['value'] = dfv['value'].agg(','.join)
                dfx['name'] = dfn['name'].agg(','.join)
                dfx['type'] = dft['type'].agg(','.join)
                # Ensures independently parsed cliques overwrite the cliques, which inherited neighbor weights
                xdf = pd.concat([dfx, cliquedf]).drop_duplicates(subset = ['entry1', 'entry2'], keep = 'last')

                if xdf[(xdf['entry1'].str.startswith('cpd:')) | (xdf['entry2'].str.startswith('cpd:'))].empty == True and xdf[(xdf['entry1'].str.startswith('undefined')) | (xdf['entry2'].str.startswith('undefined'))].empty == True:
                    # If pathway has no compounds or undefined nodes, write file
                    # Note: since this isn't a mixed pathway, "path:" nodes are excluded
                    xdf_out = xdf[(~xdf['entry1'].str.startswith('path')) & (~xdf['entry2'].str.startswith('path'))]
                    xdf_out.to_csv(output_path / '{}.tsv'.format(pathway), sep = '\t', index = False)
                else:
                    # These next series of steps are for propagating compounds and undefined nodes
                    # Uses networkx
                    G = nx.from_pandas_edgelist(xdf, source = 'entry1', target = 'entry2', edge_attr = 'name', create_using = nx.DiGraph())
                    new_edges = []
                    for node in G.nodes:
                        # Gather roots and leaflets
                        roots = [n for n, d in G.in_degree() if d == 0]
                        leaflets = [n for n, d in G.out_degree() if d == 0]
                        # Find compounds or undefined proteins that might need propagation
                        if node.startswith('cpd') or node.startswith('undefined'):
                            if node in roots or node in leaflets:
                                # Passes root nodes and leaflet nodes
                                # These do not need propagation as they terminate
                                pass
                            else:
                                out_edges = [e for e in G.out_edges(node)]
                                in_edges = [e for e in G.in_edges(node)]
                                for i in in_edges:
                                    for o in out_edges:
                                        if not i[0].startswith('cpd') and not o[1].startswith('cpd') and not i[0].startswith('undefined') and not o[1].startswith('undefined') and not i[0].startswith('path') and not o[1].startswith('path'):
                                            # Simple compound propagation removes compound between two genes
                                            # Example: hsa:xxx -> cpd:xxx -> hsa:xxx to hsa:xxx -> hsa:xxx
                                            new_edges.append([i[0], o[1], 'CPp', 'Custom', 'compound propagation'])
                                        else:
                                            # Loops through roots and leaves to find the start and end of each node that is propagated
                                            # These pathways have copious amounts of compounds and undefined genes in between the propagated genes
                                            # This loop also picks up on non-proteins that are left out if the above simple compound propagation is used
                                            for root in roots:
                                                for leaf in leaflets:
                                                    if nx.has_path(G, root, node) == True and nx.has_path(G, node, leaf) == True:
                                                        # Uses the shortest path between the root/leaf and query node
                                                        rpath = [p for p in nx.shortest_path(G, root, node)]
                                                        lpath = [p for p in nx.shortest_path(G, node, leaf)]
                                                        # Skip paths that have no proteins
                                                        if all(node.startswith('cpd') or node.startswith('undefined') or node.startswith('path') for node in rpath) == True or all(node.startswith('cpd') or node.startswith('undefined') or node.startswith('path') for node in lpath) == True:
                                                            pass
                                                        else:
                                                            # Indexes all elements in the shortest path list that is a gene
                                                            rindex = [index for index, element in enumerate(rpath) if not element.startswith('cpd') and not element.startswith('undefined') and not element.startswith('path')]
                                                            lindex = [index for index, element in enumerate(lpath) if not element.startswith('cpd') and not element.startswith('undefined') and not element.startswith('path')]
                                                            # Here we use the maximum of the root path and the minimum of the leaf path to successfully propagate compounds
                                                            # Note below example
                                                            # [mmu:x, cpd:###, cpd:x] --> max is mmu:x 
                                                            # [cpd:x, cpd:###, undefined, mmu:y, mmu:###] --> min is mmu:y
                                                            # creates edge [mmu:x, mmu:y]
                                                            new_edges.append([rpath[max(rindex)], lpath[min(lindex)], 'CPp', 'Custom', 'compound propagation'])
                    # Creates a dataframe of the new edges that are a result of compound propagation
                    new_edges_df = pd.DataFrame(new_edges, columns = df_out.columns)
                    # Concatenates the new edges with the edges from the above (cliques and original parsed edges)
                    df0 = pd.concat([xdf, new_edges_df])
                    # Drop any duplicated edges
                    df1 = df0.drop_duplicates()
                    # Removes compounds and undefined as they were propagated and no longer needed
                    df2 = df1[(~df1['entry1'].str.startswith('cpd')) & (~df1['entry2'].str.startswith('cpd')) & (~df1['entry1'].str.startswith('undefined')) & (~df1['entry2'].str.startswith('undefined')) & (~df1['entry1'].str.startswith('path')) & (~df1['entry2'].str.startswith('path'))]
                    df2.to_csv(output_path / '{}.tsv'.format(pathway), sep = '\t', index = False)
            


from setuptools import setup, find_packages

with open("README.md", "r", encoding = "utf-8") as fh:
    long_description = fh.read()
    
setup(
   name='knext',
   version='1.1.122',
   author = "Everest Uriel Castaneda",
   author_email = "Everest_Castaneda1@baylor.edu",
   description = "Kyoto Encylopedia of Genes and Genomes Markup Language File parser and converter",
   long_description = long_description,
   long_description_content_type = "text/markdown",
   url = "https://github.com/everest-castaneda/knext",
   project_urls = {
       "Bug Tracker": "https://github.com/everest-castaneda/knext/issues",
       },
   classifiers = [
       "Programming Language :: Python :: 3",
       "License :: OSI Approved :: MIT License",
       "Operating System :: OS Independent",
       ],
   packages=find_packages(),
   install_requires=[
        "networkx",
        "Click",
        "pandas",
        "typer",
        "requests",
        "pathlib",
        "pytest"
   ],
   entry_points='''
      [console_scripts]
      knext=src.main:cli
      ''',
      )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Everest Uriel Castaneda
@desc: File for converting pathway TSV files into UniProt and NCBI IDs
"""

import re
import json
import typer
import pandas as pd
from pathlib import Path
import urllib.request as request

pd.options.mode.chained_assignment = None
app = typer.Typer()

def UP(species):
    url = 'http://rest.kegg.jp/conv/%s/uniprot'
    response = request.urlopen(url % species).read().decode('utf-8')
    response = response.rstrip().rsplit('\n')
    entrez = []
    uniprot = []
    for resp in response:
        uniprot.append(resp.rsplit()[0])
        entrez.append(resp.rsplit()[1])
    d = {}
    for key, value in zip(entrez, uniprot):
        if key not in d:
            d[key] = [value]
        else:
            d[key].append(value)
    return d

def NCBI(species):
    url = 'http://rest.kegg.jp/conv/%s/ncbi-geneid'
    response = request.urlopen(url % species).read().decode('utf-8')
    response = response.rstrip().rsplit('\n')
    ncbi = []
    kegg = []
    for resp in response:
        ncbi.append(resp.rsplit()[0])
        kegg.append(resp.rsplit()[1])
    d = {}
    for key, value in zip(kegg, ncbi):
        if key not in d:
            d[key] = value
        else:
            d[key].append(value)
    return d

def convert_file(species, input_data, wd: Path, graphics = None, uniprot: bool = False, unique: bool = False):
    """
    Converts a file or folder of parsed genes or mixed pathways from KEGG IDs to 
    NCBI gene IDs or UniProt IDs. Default is NCBI gene IDs unless the
    -up/--uniprot flag is used. The -g/--graphics flag converts the graphics file
    or folder into the accompanying IDs. Do not mix/match! Use the -u/--unique 
    flag only if the input is unique, presence of terminal modifiers. 
    Use the -g/--graphics flag with folders if the input for the conversion is 
    a folder and vice-versa.
    """
    if Path(input_data).is_file():
        file = Path(input_data)
        if uniprot and unique:
            typer.echo(f'Now converting {file.name} to UniProt IDs...')
            df = pd.read_csv(file, delimiter = '\t')
            conv = UP(species)
            # Extract the terminal modifiers and create a new column
            # This enables the re-addition of the modifiers at the
            # end of the function.
            df['match1'] = df['entry1'].str.extract(r'(-[0-9]+)')
            df['match2'] = df['entry2'].str.extract(r'(-[0-9]+)')
            # Remove the terminal modifier so that the IDs map properly
            # to the KEGG API call
            df['entry1'] = df['entry1'].str.replace(r'(-[0-9]+)', '', regex=True)
            df['entry2'] = df['entry2'].str.replace(r'(-[0-9]+)', '', regex=True)
            # Map to convert KEGG IDs to UniProt IDs. Note lists are returned
            # for some conversions.
            df['entry1'] = df['entry1'].map(conv)
            df['entry2'] = df['entry2'].map(conv)
            # Drop any IDs that did not convert.
            df1 = df.dropna()
            # Joins the comma seperated list to a string to avoid iterrating
            df1['entry1'] = [','.join(map(str, l)) for l in df1['entry1']]
            df1['entry2'] = [','.join(map(str, l)) for l in df1['entry2']]
            # Quick replace of all up: modifiers at the beginning of each string
            df1 = df1.replace(to_replace = 'up:', value = '', regex = True)
            # Quick seperation back into a list.
            df1['entry1'] = df1['entry1'].apply(lambda x: re.findall(r'[a-zA-z0-9]+', x))
            df1['entry2'] = df1['entry2'].apply(lambda x: re.findall(r'[a-zA-z0-9]+', x))
            df_out = df1.explode('entry1', ignore_index = True).explode('entry2', ignore_index = True)
            df_out['entry1'] = df_out['entry1'] + df_out['match1']
            df_out['entry2'] = df_out['entry2'] + df_out['match2']
            df_out.drop(['match1', 'match2'], axis = 1, inplace = True)
            df_out.to_csv(wd / 'up_{}'.format(file.name), sep = '\t', index = False)
            if graphics != None and Path(graphics).is_file():
                typer.echo(f'Graphics file given! Now converting {Path(graphics).name} to UniProt IDs...')
                pos = open(graphics)
                d = json.loads(pos.read())
                up_dict = {}
                for key, items in d.items():
                    pattern = re.search(r'(-[0-9]+)', key)
                    new_keys = re.sub(r'(-[0-9]+)', '', key)
                    try:
                        uplist = conv[new_keys]
                        new_up = [up + pattern.group() for up in uplist]
                        up_dict[str(new_up)] = items
                        for up in new_up:
                            up_dict[up.replace('up:', '')] = items
                    except KeyError:
                        pass
                with open(wd / 'up_{}'.format(Path(graphics).name), 'w') as outfile:
                    outfile.write(json.dumps(up_dict))
        elif not uniprot and unique:
            typer.echo(f'Now converting {file.name} to NCBI IDs...')
            df = pd.read_csv(file, delimiter = '\t')
            conversion = NCBI(species)
            df['match1'] = df['entry1'].str.extract(r'(-[0-9]+)')
            df['match2'] = df['entry2'].str.extract(r'(-[0-9]+)')
            df['entry1'] = df['entry1'].str.replace(r'(-[0-9]+)', '', regex=True)
            df['entry2'] = df['entry2'].str.replace(r'(-[0-9]+)', '', regex=True)
            df['entry1'] = df['entry1'].map(conversion)
            df['entry2'] = df['entry2'].map(conversion)
            df1 = df.dropna()
            df1['entry1'] = df1['entry1'].str.replace('ncbi-geneid:', '') + df1['match1']
            df1['entry2'] = df1['entry2'].str.replace('ncbi-geneid:', '') + df1['match2']
            df_out = df1.drop(['match1', 'match2'], axis = 1) 
            df_out.to_csv(wd / 'ncbi_{}'.format(file.name), sep = '\t', index = False)
            if graphics != None and Path(graphics).is_file():
                typer.echo(f'Graphics file given. Now converting {Path(graphics).name}...')
                pos = open(graphics)
                d = json.loads(pos.read())
                ncbi_dict = {}
                for key, items in d.items():
                    pattern = re.search(r'(-[0-9]+)', key)
                    new_keys = re.sub(r'(-[0-9]+)', '', key)
                    try:
                        ncbi_dict[conversion[new_keys].replace('ncbi-geneid:', '') + pattern.group()] = items
                    except KeyError:
                        pass
                with open(wd / 'ncbi_{}'.format(Path(graphics).name), 'w') as outfile:
                    outfile.write(json.dumps(ncbi_dict))
        elif uniprot and not unique:
            typer.echo(f'Now converting {file.name} to UniProt IDs...')
            df = pd.read_csv(file, delimiter = '\t')
            conv = UP(species)
            df['entry1'] = df['entry1'].map(conv)
            df['entry2'] = df['entry2'].map(conv)
            df=df.dropna()
            df_out = df.explode('entry1', ignore_index = True).explode('entry2', ignore_index = True)
            df_out['entry1'] = df_out['entry1'].str.replace('up:', '')
            df_out['entry2'] = df_out['entry2'].str.replace('up:', '')
            df_out.to_csv(wd / 'up_{}'.format(file.name), sep = '\t', index = False)
            if graphics != None and Path(graphics).is_file():
                typer.echo(f'Graphics file given! Now converting {graphics.name} to UniProt IDs...')
                pos = open(graphics)
                d = json.loads(pos.read())
                up_dict = {}
                for key, items in d.items():
                    try:
                        uplist = conv[key]
                        for up in uplist:
                            up_dict[up.replace('up:', '')] = items
                    except KeyError:
                        pass
                with open(wd / 'up_%s' % Path(graphics).name, 'w') as outfile:
                    outfile.write(json.dumps(up_dict))
        elif not uniprot and not unique:
            typer.echo(f'Now converting {file.name} to NCBI IDs...')
            df = pd.read_csv(file, delimiter = '\t')
            conversion = NCBI(species)
            df['entry1'] = df['entry1'].map(conversion).str.replace('ncbi-geneid:', '')
            df['entry2'] = df['entry2'].map(conversion).str.replace('ncbi-geneid:', '')
            df1 = df.dropna()
            df_out.to_csv(wd / 'ncbi_{}'.format(file.name), sep = '\t', index = False)
            if graphics != None and Path(graphics).is_file():
                typer.echo(f'Graphics file given! Now converting {graphics} to NCBI IDs...')
                pos = open(graphics)
                d = json.loads(pos.read())
                ncbi_dict = {}
                for key, items in d.items():
                    try:
                        ncbi_dict[conversion[key].replace('ncbi-geneid:', '')] = items
                    except KeyError:
                        pass
                with open(wd / 'ncbi_%s' % Path(graphics).name, 'w') as outfile:
                    outfile.write(json.dumps(ncbi_dict))

def convert_folder(species, input_data, wd: Path, graphics = None, uniprot: bool = False, unique: bool = False):
    if uniprot and not unique:
        output_path = wd / 'up_{}'.format(species)
        output_path.mkdir(exist_ok = True)
        for file in Path(input_data).glob('*.tsv'):
            typer.echo(f'Now converting {file.name} to UniProt IDs...')
            df = pd.read_csv(file, delimiter = '\t')
            conv = UP(species)
            df['entry1'] = df['entry1'].map(conv)
            df['entry2'] = df['entry2'].map(conv)
            df = df.dropna()
            df_out = df.explode('entry1', ignore_index = True).explode('entry2', ignore_index = True)
            df_out['entry1'] = df_out['entry1'].str.replace('up:', '')
            df_out['entry2'] = df_out['entry2'].str.replace('up:', '')
            df_out.to_csv(output_path / 'up_{}'.format(file.name), sep = '\t', index = False)
        if graphics != None and Path(graphics).is_dir():
            for f in Path(graphics).glob('*.txt'):
                typer.echo(f'Graphics folder given! Now converting {f.name} to UniProt IDs...')
                pos = open(f)
                d = json.loads(pos.read())
                up_dict = {}
                for key, items in d.items():
                    try:
                        uplist = conv[key]
                        for up in uplist:
                            up_dict[up.replace('up:', '')] = items
                    except KeyError:
                        pass
                with open(output_path /'up_{}'.format(f.name), 'w') as outfile:
                    outfile.write(json.dumps(up_dict))
    elif not uniprot and not unique:
        output_path = wd / 'ncbi_{}'.format(species)
        output_path.mkdir(exist_ok = True)
        for file in Path(input_data).glob('*.tsv'):
            typer.echo(f'Now converting {file.name} to NCBI IDs...')
            df = pd.read_csv(file, delimiter = '\t')
            conversion = NCBI(species)
            df['entry1'] = df['entry1'].map(conversion).str.replace('ncbi-geneid:', '')
            df['entry2'] = df['entry2'].map(conversion).str.replace('ncbi-geneid:', '')
            df1 = df.dropna()
            df1.to_csv(output_path / 'ncbi_{}'.format(file.name), sep = '\t', index = False)
        if graphics != None and Path(graphics).is_dir():
            for f in Path(graphics).glob('*.txt'):
                typer.echo(f'Graphics folder given! Now converting {f.name} to NCBI IDs...')
                pos = open(f)
                d = json.loads(pos.read())
                ncbi_dict = {}
                for key, items in d.items():
                    try:
                        ncbi_dict[conversion[key].replace('ncbi-geneid:', '')] = items
                    except KeyError:
                        pass
                with open(output_path / 'ncbi_{}'.format(f.name), 'w') as outfile:
                    outfile.write(json.dumps(ncbi_dict))
    elif uniprot and unique:
        output_path = wd / 'up_{}'.format(species)
        output_path.mkdir(exist_ok = True)
        for file in Path(input_data).glob('*.tsv'):
            typer.echo(f'Now converting {file.name} to UniProt IDs...')
            df = pd.read_csv(file, delimiter = '\t')
            conv = UP(species)
            # Extract the terminal modifiers and create a new column
            # This enables the readdition of the modifiers at the
            # end of the function.
            df['match1'] = df['entry1'].str.extract('(-[0-9]+)')
            df['match2'] = df['entry2'].str.extract('(-[0-9]+)')
            # Remove the terminal modifier so that the IDs map properly
            # to the KEGG API call
            df['entry1'] = df['entry1'].str.replace(r'(-[0-9]+)', '', regex=True)
            df['entry2'] = df['entry2'].str.replace(r'(-[0-9]+)', '', regex=True)
            # Map to convert KEGG IDs to UniProt IDs. Note lists are returned
            # for some conversions.
            df['entry1'] = df['entry1'].map(conv)
            df['entry2'] = df['entry2'].map(conv)
            # Drop any IDs that did not convert.
            # This avoids errors in the subsequent entry that creates lists
            df1 = df.dropna()
            # Joins the comma seperated list to a string to avoid iterrating
            df1['entry1'] = [','.join(map(str, l)) for l in df1['entry1']]
            df1['entry2'] = [','.join(map(str, l)) for l in df1['entry2']]
            # Quick replace of all up: modifiers at the beginning of each string
            df1 = df1.replace(to_replace = 'up:', value = '', regex = True)
            # Quick seperation back into a list.
            df1['entry1'] = df1['entry1'].apply(lambda x: re.findall(r'[a-zA-z0-9]+', x))
            df1['entry2'] = df1['entry2'].apply(lambda x: re.findall(r'[a-zA-z0-9]+', x))
            # Individualize each entry from a list
            df_out = df1.explode('entry1', ignore_index = True).explode('entry2', ignore_index = True)
            # Add back the unique modifiers
            df_out['entry1'] = df_out['entry1'] + df_out['match1']
            df_out['entry2'] = df_out['entry2'] + df_out['match2']
            # Drop the unique modifiers columns
            df_out.drop(['match1', 'match2'], axis = 1, inplace = True)
            df_out.to_csv(output_path / 'up_{}'.format(file.name), sep = '\t', index = False)
        if graphics != None and Path(graphics).is_dir():
            for f in Path(graphics).glob('*.txt'):
                typer.echo(f'Graphics folder given! Now converting {f.name}...')
                pos = open(f)
                d = json.loads(pos.read())
                up_dict = {}
                for key, items in d.items():
                    pattern = re.search(r'(-[0-9]+)', key)
                    new_keys = re.sub(r'(-[0-9]+)', '', key)
                    try:
                        uplist = conv[new_keys]
                        new_up = [up + pattern.group() for up in uplist]
                        up_dict[str(new_up)] = items
                        for up in new_up:
                            up_dict[up.replace('up:', '')] = items
                    except KeyError:
                        pass
                with open(output_path / 'up_{}'.format(f.name), 'w') as outfile:
                    outfile.write(json.dumps(up_dict))
    elif not uniprot and unique:
        output_path = wd / 'ncbi_{}'.format(species)
        output_path.mkdir(exist_ok = True)
        for file in Path(input_data).glob('*.tsv'):
            typer.echo(f'Now converting {file.name} to NCBI IDs...')
            df = pd.read_csv(file, delimiter = '\t')
            conversion = NCBI(species)
            df['match1'] = df['entry1'].str.extract(r'(-[0-9]+)')
            df['match2'] = df['entry2'].str.extract(r'(-[0-9]+)')
            df['entry1'] = df['entry1'].str.replace(r'(-[0-9]+)', '', regex=True)
            df['entry2'] = df['entry2'].str.replace(r'(-[0-9]+)', '', regex=True)
            df['entry1'] = df['entry1'].map(conversion)
            df['entry2'] = df['entry2'].map(conversion)
            df1 = df.dropna()
            df1['entry1'] = df1['entry1'].str.replace('ncbi-geneid:', '') + df1['match1']
            df1['entry2'] = df1['entry2'].str.replace('ncbi-geneid:', '') + df1['match2']
            df2 = df1.drop(['match1', 'match2'], axis = 1) 
            df2.to_csv(output_path / 'ncbi_{}'.format(file.name), sep = '\t', index = False)
        if graphics != None and Path(graphics).is_dir():
            for f in Path(graphics).glob('*.txt'):
                typer.echo(f'Graphics folder given! Now converting {f.name} to NCBI IDs...')
                pos = open(f)
                d = json.loads(pos.read())
                ncbi_dict = {}
                for key, items in d.items():
                    pattern = re.search(r'(-[0-9]+)', key)
                    new_keys = re.sub(r'(-[0-9]+)', '', key)
                    try:
                        ncbi_dict[conversion[new_keys].replace('ncbi-geneid:', '') + pattern.group()] = items
                    except KeyError:
                        pass
                with open(output_path / 'ncbi_{}'.format(f.name), 'w') as outfile:
                    outfile.write(json.dumps(ncbi_dict))
                    
                        
                        
            
    
    
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb  1 09:55:10 2023

@author: everest_castaneda1
"""

import xml.etree.ElementTree as ET
import json
import pandas as pd
import networkx as nx

def test_parser():
    tree = ET.parse('data/hsa00232.xml')
    root = tree.getroot()
    e1 = []
    e2 = []
    n = []
    v = []
    for relation in root.findall('relation'):
        e1.append(relation.get('entry1'))
        e2.append(relation.get('entry2'))
        for subtype in relation.findall('subtype'):
            n.append(subtype.get('name'))
            v.append(subtype.get('value'))
           
    d = []
    for relation in root.findall('relation'):
        for subtype in relation:
            d1=relation.attrib
            d2=subtype.attrib
            d3=json.dumps(d1),json.dumps(d2)
            d.append(d3)    

    edgelist = []
    for line in d:
        x = line[0].replace("{","").replace("}","").replace('"entry1":',"")\
            .replace('"entry2":',"").replace('"type":',"").replace('"name":',"")\
            .replace('"value":',"").replace('"','').replace(',',"\t").replace(' ', '')
        y = line[1].replace("{","").replace("}","").replace('"name":',"")\
            .replace('"',"").replace('value:',"").replace(",",'\t').replace(' ', '')
        edgelist.append(x+"\t"+y)

    df = pd.DataFrame(edgelist)
    df = df[0].str.split("\t", expand=True).rename({0: 'entry1',1: 'entry2',
                                                  2: 'types', 3:'name',
                                                  4: 'value'}, axis='columns')
    G = nx.from_pandas_edgelist(df, source = 'entry1', target = 'entry2')
    assert G.edges